#!/usr/bin/perl

# AC3D parser and converter to Horde3D format
# by Niek Albers

my $VERSION = 0.6;

use strict;
no strict 'refs';

#no warnings;
use Data::Dumper;
use Carp;
use File::Copy;

my $AC3D_FILE_VERSION = 11;

my $normal = grep { $_ eq '-n' } @ARGV;
my $scene_basename;

my $ac3dfile;
my $scenefile;
my $geofile;
my @VERTICES;
my @INDICES;
my @NORMALS;
my @TANGENTS;
my @BITANGENTS;
my @TCOORDS0;
my @TCOORDS1;

my $index_offset   = 0;
my $batch_offset   = 0;
my $world          = {};
my $material_files = {};
my $last_name;

# my $vertice_count = 0;

my $TOKENIZER = {
    'MATERIAL' => \&parse_material,
    'OBJECT'   => \&parse_object,
    'kids'     => \&parse_kids,
    'name'     => \&parse_name,
    'texture'  => \&parse_value,
    'crease'   => \&parse_value,
    'numvert'  => \&parse_numvert,
    'numsurf'  => \&parse_numsurf,
    'SURF'     => \&parse_surf,
    'mat'      => \&parse_mat,
    'refs'     => \&parse_refs,
    'data'     => \&parse_value,
    'texrep'   => \&parse_texrep,
    'rot'      => \&parse_rot,
    'loc'      => \&parse_loc,
    'url'      => \&parse_value,
};
eval { main(@ARGV) };
die($@) if ($@);

sub basename {
    my $path = shift;
    $path =~ s/.*[\/\\]//;
    return [ split( /\./, $path ) ]->[0];
}

sub filename {
    my $path = shift;
    $path =~ s/.*[\/\\]//;
    return $path;
}

sub parse_line {
    my $line = <$ac3dfile>;
    return if ( !$line );
    my @items = ( $line =~ /(".*?"|\S+)/g );
    map { s/(^"|"$)//g } @items;
    return \@items;

}

sub parse_header {
    my $items = shift;
    $items->[0] =~ m/^AC3D(\w)/;
    die "not minimally an AC3D version $AC3D_FILE_VERSION file\n"
      if ( hex($1) < $AC3D_FILE_VERSION );
}

sub parse_roc {
    my $items = shift;
    return [
        [ $items->[1], $items->[2], $items->[3] ],
        [ $items->[4], $items->[5], $items->[6] ],
        [ $items->[7], $items->[8], $items->[9] ]
    ];
}

sub parse_value {
    my $items = shift;
    return $items->[1];
}

sub parse_name {
    my $items = shift;
    $last_name = $items->[1];
    return $items->[1];
}

sub parse_texrep {
    my $items = shift;

    return { x => $items->[1], y => $items->[2] };
}

sub parse_surf {
    my $items = shift;

    my $param = hex( $items->[1] );
    my $type  = $param & 0xF;
    my $types = { 0 => 'polygon', 1 => 'closedline', 2 => 'line' };

    my $shading_sides = $param >> 4;

    my $shading = $shading_sides & 0x1 ? 'smooth' : 'flat';
    my $sides   = $shading_sides & 0x2 ? 2        : 1;

    my $result = {
        type    => $types->{$type},
        shading => $shading,
        sides   => $sides,
    };

    return $result;
}

sub parse_mat {
    my $items = shift;
    my $index = $items->[1];

    return $world->{MATERIAL}->[$index];

}

sub parse_loc {
    my $items = shift;
    return vec3( $items->[1], $items->[2], $items->[3] );
}

sub parse_numvert {
    my $items    = shift;
    my $numverts = $items->[1];
    my @vertices;
    for ( 1 .. $numverts ) {
        my $items = parse_line();
        push @vertices, vec3( $items->[0], $items->[1], $items->[2] );
    }
    return \@vertices;
}

sub parse_refs {
    my $items   = shift;
    my $numrefs = $items->[1];
    die("model is not triangulated: $last_name\n") if ( $numrefs != 3 );
    my @indices;
    for ( 1 .. $numrefs ) {
        my $items = parse_line();
        push @indices,
          { index => $items->[0], u => $items->[1], v => $items->[2] };
    }
    return \@indices;
}

sub parse_surface {
    my $items   = shift;
    my $surface = {};
    while ( my $items = parse_line() ) {
        my $opcode = $items->[0];
        if ( exists( $TOKENIZER->{$opcode} ) ) {
            $surface->{$opcode} = $TOKENIZER->{$opcode}($items);
        }
        last if ( $surface->{refs} );
    }
    return $surface;
}

sub parse_numsurf {
    my $items    = shift;
    my $numsurfs = $items->[1];
    my @surfaces;
    for ( 1 .. $numsurfs ) {
        push @surfaces, parse_surface($items);
    }
    return \@surfaces;
}

sub parse_kids {
    my $items  = shift;
    my $amount = $items->[1];
    my @objects;
    for ( 1 .. $amount ) {
        my $items = parse_line();
        push @objects, parse_object($items);
    }

    return \@objects;
}

sub parse_object {
    my $items = shift;

    my $type   = $items->[1];
    my $object = {};
    my $error;
    $object->{type} = $type;
    while ( my $items = parse_line() ) {
        my $opcode = $items->[0];
        if ( exists( $TOKENIZER->{$opcode} ) ) {
            $object->{$opcode} = $TOKENIZER->{$opcode}($items);
            $error = $@ if ($@);
        }
        last if ( $opcode eq 'kids' );
    }
    warn "parsed object: $object->{name}, type: $type, vertices: "
      . int( @{ $object->{numvert} } ) . "\n";
    return $object;
}

sub parse_material {
    my $items = shift;
    return {
        name  => $items->[1],
        rgb   => [ $items->[3], $items->[4], $items->[5] ],
        amb   => [ $items->[7], $items->[8], $items->[9] ],
        emis  => [ $items->[11], $items->[12], $items->[13] ],
        spec  => [ $items->[15], $items->[16], $items->[17] ],
        shi   => $items->[19],
        trans => $items->[21],
    };
}

sub parse {

    while ( my $items = parse_line() ) {
        my $command = $items->[0];
        if ( exists( $TOKENIZER->{$command} ) ) {
            push @{ $world->{$command} }, $TOKENIZER->{$command}($items);
        }
    }
    return $world;
}

sub write_material {
    my $object           = shift;
    my $texture          = $object->{texture} || $scene_basename || 'standard';
    my $texture_basename = basename($texture);
    my $use_normal       = $normal;

    my $material_filename = "$scene_basename/$texture_basename.material.xml";

    return if ( $material_files->{$material_filename} );

    warn "writing to: $material_filename\n";
    my $materialfile;
    mkdir("$scene_basename");
    open( $materialfile, ">$material_filename" )
      || die("Cannot write to $material_filename\n");
    print $materialfile qq~<Material>\n~;
    my $texture_filename = filename($texture);

    warn
      qq~copying "$texture_filename" to "$scene_basename/$texture_filename"\n~;
    copy( $texture_filename, "$scene_basename/$texture_filename" ) || die("$!\n");

    my $texture_normal_filename = $texture_filename;
    $texture_normal_filename =~ s/\.(\w\w\w)$/_normal.$1/;

    if ( !-e "$texture_normal_filename" ) {
        $use_normal = 0;
    }
    else {
        warn
qq~copying "$texture_normal_filename" to "$scene_basename/$texture_normal_filename"\n~;
        copy( $texture_normal_filename, "$scene_basename/$texture_normal_filename" )
          || die("$!\n");
    }

    print $materialfile qq~<Shader source="shaders/model.shader" />\n~;

    if ($texture) {
        if ($use_normal) {
            print $materialfile qq~<ShaderFlag name="_F02_NormalMapping" />\n~;
        }
        print $materialfile
qq~<Sampler name="albedoMap" map="$scene_basename/$texture_filename" />\n~;
        if ($use_normal) {    # add normal mapped texture
            print $materialfile
qq~<Sampler name="normalMap" map="$scene_basename/$texture_normal_filename" />\n~;
        }
    }
    print $materialfile qq~</Material>\n~;
    close($materialfile);
    $material_files->{$material_filename} = 1;
}

# add vertices for indices that point to the same vertex (full unweld)
sub unweld_object {
    my $object = shift;

    print STDERR "unwelding: $object->{name} ";

    my @tcoords;
    my %found;

    foreach my $surface ( @{ $object->{numsurf} } ) {
        foreach my $ref ( @{ $surface->{refs} } ) {
            if ( $found{ $ref->{index} } ) {

                # add new vertex
                push @{ $object->{numvert} },
                  $object->{numvert}->[ $ref->{index} ];
                $ref->{index} = int( @{ $object->{numvert} } ) - 1;
            }
            $found{ $ref->{index} } = 1;
            $tcoords[ $ref->{index} ] = $ref;
        }
    }
    print STDERR "vertices: " . int( @{ $object->{numvert} } ) . "\n";

    $object->{tcoords} = \@tcoords;
}

# remove vertices that have the same coordinates, texture coordinates and normals
sub weld_object {
    my $object = shift;

    print STDERR "welding: $object->{name} ";

    my @vertices;
    my @normals;
    my @tangents;
    my @bitangents;
    my @tcoords;
    my %found;
    my $index = 0;
    foreach my $surface ( @{ $object->{numsurf} } ) {

        foreach my $ref ( @{ $surface->{refs} } ) {
            my $vertex = $object->{numvert}->[ $ref->{index} ];
            my $normal = $object->{normals}->[ $ref->{index} ];
            my $tcoord = $object->{tcoords}->[ $ref->{index} ];

            my $tangent   = $object->{tangents}->[ $ref->{index} ];
            my $bitangent = $object->{bitangents}->[ $ref->{index} ];

            if (
                !exists(
                    $found{ $vertex->as_string }->{ $normal->as_string }
                      ->{ $tcoord->{u} . ',' . $tcoord->{v} }
                )
              )
            {
                $found{ $vertex->as_string }->{ $normal->as_string }
                  ->{ $tcoord->{u} . ',' . $tcoord->{v} } = $index;

                $vertices[$index]   = $vertex;
                $normals[$index]    = $normal;
                $tcoords[$index]    = $tcoord;
                $tangents[$index]   = $tangent;
                $bitangents[$index] = $bitangent;

                $index++;
            }
            $ref->{index} =
              $found{ $vertex->as_string }->{ $normal->as_string }
              ->{ $tcoord->{u} . ',' . $tcoord->{v} };
        }
    }

    $object->{numvert}    = \@vertices;
    $object->{normals}    = \@normals;
    $object->{tangents}   = \@tangents;
    $object->{bitangents} = \@bitangents;
    $object->{tcoords}    = \@tcoords;

    print STDERR "vertices: " . int( @{ $object->{numvert} } ) . "\n";
}

# find lowest and highest vertex index
sub get_min_max_count_index {
    my $object      = shift;
    my $min_index   = 0xffffffff;
    my $max_index   = 0;
    my $index_count = 0;

    foreach my $surface ( @{ $object->{numsurf} } ) {

        foreach my $ref ( @{ $surface->{refs} } ) {
            if ( $max_index < $ref->{index} ) {
                $max_index = $ref->{index};
            }
            if ( $min_index > $ref->{index} ) {
                $min_index = $ref->{index};
            }
            $index_count++;
        }
    }
    return ( $min_index, $max_index, $index_count );
}

sub write_object {
    my $object = shift;

    my $tag    = '';
    my $params = {};

    if ( $object->{loc} ) {
        $params->{tx} = $object->{loc}->x;
        $params->{ty} = $object->{loc}->y;
        $params->{tz} = $object->{loc}->z;
    }

    if ( $object->{type} eq 'world' ) {
        $tag = 'Group';
        $params->{name} = 'world';
    }
    elsif ( $object->{type} eq 'group' ) {
        $tag                = 'Model';
        $params->{name}     = $object->{name};
        $params->{geometry} = "$scene_basename/$scene_basename.geo";
        $object->{in_model} = 1;
    }
    elsif ( $object->{type} eq 'poly' ) {

        if ( !$object->{in_model} )
        {    # whoops not in a model yet, create one first
            my %wrapper = %$object;
            $wrapper{kids} = [$object];
            $wrapper{type} = 'group';
            delete $object->{loc};    # set location for model in this case
            write_object( \%wrapper );
            return;
        }
        else {

            $tag = 'Mesh';

            unweld_object($object);
            calc_vertex_normals($object);
            weld_object($object);

            #            warn Dumper($object);

            my $material = basename( $object->{texture} ) || 'standard';
            $params->{material} = "$scene_basename/$material.material.xml";
            $params->{name}     = $object->{name};

            my ( $min_index, $max_index, $index_count ) =
              get_min_max_count_index($object);

            $params->{vertRStart} = $min_index + $index_offset;
            $params->{vertREnd}   = $max_index + $index_offset;

            $params->{batchCount} = $index_count;
            $params->{batchStart} = $batch_offset;

            write_material($object);

            push @VERTICES,   @{ $object->{numvert} };
            push @NORMALS,    @{ $object->{normals} };
            push @TANGENTS,   @{ $object->{tangents} };
            push @BITANGENTS, @{ $object->{bitangents} };
            push @TCOORDS0,   @{ $object->{tcoords} };

            foreach my $surface ( @{ $object->{numsurf} } ) {
                foreach my $ref ( @{ $surface->{refs} } ) {
                    push @INDICES, $ref->{index} + $index_offset;
                }
            }
            $index_offset += int( @{ $object->{numvert} } );
            $batch_offset += $index_count;

        }
    }

    my $attributes = '';
    foreach my $key ( sort keys(%$params) ) {
        $attributes .= qq~ $key="$params->{$key}"~;
    }
    print $scenefile qq~<$tag$attributes>\n~;
    foreach my $kid ( @{ $object->{kids} } ) {
        $kid->{in_model} = $object->{in_model};
        write_object($kid);
    }
    print $scenefile qq~</$tag>\n~;
    delete $object->{kids};
}

sub write_geo {

    my $filename = "$scene_basename/$scene_basename.geo";
    open( $geofile, ">$filename" )
      || die("cannot write to $filename\n");
    binmode($geofile);

    warn "writing to: $filename\n";

    print $geofile "H3DG" . pack( 'I', 5 );    # horde file version
    print $geofile pack( 'I', 1 );             # numjoints (0+1)
    print $geofile
      pack( 'f16', 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1 )
      ;                                        # identity matrix

    print $geofile pack( 'I', 6 );                 # num vertex streams
    print $geofile pack( 'I', int(@VERTICES) );    # num vertices

    # Vertices - positions magic = 0
    print $geofile pack( 'I', 0 );

    # streamElementSize 12
    print $geofile pack( 'I', 12 );

    foreach my $vertex (@VERTICES) {
        print $geofile pack( 'f3', $vertex->x, $vertex->y, $vertex->z );
    }

    # Normals - positions magic = 1
    print $geofile pack( 'I', 1 );

    # streamElementSize 6
    print $geofile pack( 'I', 6 );

    foreach my $normal (@NORMALS) {
        print $geofile pack( 's3',
            round( $normal->x * 32767 ),
            round( $normal->y * 32767 ),
            round( $normal->z * 32767 ) );
    }

    # Tangents - positions magic = 2
    print $geofile pack( 'I', 2 );

    # streamElementSize 6
    print $geofile pack( 'I', 6 );

    foreach my $tangent (@TANGENTS) {
        print $geofile pack( 's3',
            round( $tangent->x * 32767 ),
            round( $tangent->y * 32767 ),
            round( $tangent->z * 32767 ) );
    }

    # BiTangents - positions magic = 3
    print $geofile pack( 'I', 3 );

    # streamElementSize 6
    print $geofile pack( 'I', 6 );

    foreach my $bitangent (@BITANGENTS) {
        print $geofile pack( 's3',
            round( $bitangent->x * 32767 ),
            round( $bitangent->y * 32767 ),
            round( $bitangent->z * 32767 ) );
    }

    # texture coordinates set 0 - positions magic = 6
    print $geofile pack( 'I', 6 );

    # streamElementSize 8
    print $geofile pack( 'I', 8 );

    foreach my $tcoord (@TCOORDS0) {
        print $geofile pack( 'f2', $tcoord->{u}, $tcoord->{v} );
    }

    # for now
    @TCOORDS1 = @TCOORDS0;

    # texture coordinates set 1 - positions magic = 7
    print $geofile pack( 'I', 7 );

    # streamElementSize 8
    print $geofile pack( 'I', 8 );
    foreach my $tcoord (@TCOORDS1) {
        print $geofile pack( 'f2', 0, 0 );

        #        print $geofile pack( 'f2', $tcoord->{u}, $tcoord->{v} );
    }

    print $geofile pack( 'I', int(@INDICES) );
    foreach my $index (@INDICES) {
        print $geofile pack( 'I', $index );
    }

    # Morph targets = 0
    print $geofile pack( 'I', 0 );
    close($geofile);

}

sub write_scene {
    my $tree = shift;

    mkdir($scene_basename);
    my $scene_filename = "$scene_basename/$scene_basename.scene.xml";
    warn "writing to: $scene_filename\n";
    open( $scenefile, ">$scene_filename" )
      || die("Cannot write to $scene_filename\n");

    write_object( $tree->{OBJECT}->[0] );
    close($scenefile);

}

sub uv_sub {
    my $uv1 = shift;
    my $uv2 = shift;

    return {
        u => $uv1->{u} - $uv2->{u},
        v => $uv1->{v} - $uv2->{v}
    };
}

# find vertices with same coordinates, this is for normal smoothing, return of list of matching indices
sub find_same_vertices {
    my $object      = shift;
    my $indices_map = shift;
    my $index       = shift;

    my $vertex  = $object->{numvert}->[$index];
    my @indices = @{ $indices_map->{ $vertex->as_string } };
    push( @indices, $index ) if ( !grep { $_ == $index } @indices );
    return @indices;
}

# create map of indices with same vertices
sub make_indices_map {
    my $object = shift;

    my %indices_map;

    foreach my $surface ( @{ $object->{numsurf} } ) {

        # only average out smooth vertex normals
        next if ( $surface->{SURF}->{shading} ne 'smooth' );
        foreach my $ref ( @{ $surface->{refs} } ) {

            my $v = $object->{numvert}->[ $ref->{index} ];
            push @{ $indices_map{ $v->as_string } }, $ref->{index};
        }
    }

    #    warn Dumper(\%indices_map);

    return \%indices_map;
}

sub calc_vertex_normals {
    my $object = shift;

    warn "calculating normals: $object->{name}\n";

    my @normals;
    my @tangents;
    my @bitangents;

    my $indices_map = make_indices_map($object);

    foreach my $surface ( @{ $object->{numsurf} } ) {

        my $index0 = $surface->{refs}->[0]->{index};
        my $index1 = $surface->{refs}->[1]->{index};
        my $index2 = $surface->{refs}->[2]->{index};

        my $v0 = $object->{numvert}->[$index0];
        my $v1 = $object->{numvert}->[$index1];
        my $v2 = $object->{numvert}->[$index2];

        my $d0 = $v1 - $v0;
        my $d1 = $v2 - $v0;
        my $v  = ( $d0 x $d1 )->normalized;

        #
        #  Calculate the normal for this face and add it
        #  to the running sum of normals for each of the
        #  three vertices involved
        #

        # and now tangents and bitangents
        my $Edge0uv = uv_sub( $surface->{refs}->[1], $surface->{refs}->[0] );
        my $Edge1uv = uv_sub( $surface->{refs}->[2], $surface->{refs}->[0] );

        my $cp = $Edge0uv->{u} * $Edge1uv->{v} - $Edge1uv->{u} * $Edge0uv->{v};
        my $r  = 0;
        if ( $cp != 0 ) {
            $r = 1.0 / $cp;
        }

        my $tangent   = ( $d0 * $Edge1uv->{v} - $d1 * $Edge0uv->{v} ) * $r;
        my $bitangent = ( $d1 * $Edge0uv->{u} - $d0 * $Edge1uv->{u} ) * $r;

        my @indices;
        push @indices, find_same_vertices( $object, $indices_map, $index0 );
        push @indices, find_same_vertices( $object, $indices_map, $index1 );
        push @indices, find_same_vertices( $object, $indices_map, $index2 );
        foreach my $index (@indices) {
            $normals[$index]    ||= vec3();
            $tangents[$index]   ||= vec3();
            $bitangents[$index] ||= vec3();

            $normals[$index]    += $v;
            $tangents[$index]   += $tangent;
            $bitangents[$index] += $bitangent;
        }

    }

    #  Normalize and fixup the vertex normal vectors, tangents en bitangents
    my $numVerts = int( @{ $object->{numvert} } );
    for ( my $j = 0 ; $j < $numVerts ; $j++ ) {
        my $n = $normals[$j]->normalized;
        my $t = $tangents[$j];

        # orthogonalize
        $tangents[$j] = ( $t - $n * ( $n * $t ) )->normalized;
        $normals[$j] = $n;

        if ( ( $n x $t ) * $bitangents[$j] < 0 ) {
            $bitangents[$j] = ( ( $n * -1 ) x $t )->normalized;
        }
        else {
            $bitangents[$j] = ( $n x $t )->normalized;
        }
    }

    $object->{normals}    = \@normals;
    $object->{tangents}   = \@tangents;
    $object->{bitangents} = \@bitangents;
}

sub round {
    my $number = shift;
    return int( $number + .5 * ( $number <=> 0 ) );
}

sub main {
    warn
"AC3D2Horde3D $VERSION - Convert AC3D ac files to Horde3D geo,scene and material files.\n";
    my $source = pop
      || die
"usage: $0 [-n] <file.ac>\n -n: add texture_normal.ext to materials and use normal shader.\n";
    open( $ac3dfile, "<$source" ) || die "cannot open $source\n";
    $scene_basename = basename($source);
    my $header = parse_line();
    parse_header($header);
    my $tree = parse();

    write_scene($tree);

    #    warn Dumper($tree);
    write_geo();
}

# simple vector3 class
sub vec3 { vec3->new(@_) }

package vec3;

sub new {
    my $caller = shift;
    my $class  = ref($caller) || $caller;
    my $self   = [ map( { defined($_) ? $_ : 0 } @_[ 0, 1, 2 ] ) ];
    bless( $self, $class );
    return ($self);
}

sub x { shift->[0] }
sub y { shift->[1] }
sub z { shift->[2] }

use overload '+' => sub {
    my ( $v1, $v2 ) = @_;
    vec3->new( $v1->x + $v2->x, $v1->y + $v2->y, $v1->z + $v2->z );
  },
  '==' => sub {
    my ( $v1, $v2 ) = @_;
    return ( $v1->x == $v2->x && $v1->y == $v2->y && $v1->z == $v2->z );
  },
  '-' => sub {
    my ( $v1, $v2 ) = @_;
    vec3->new( $v1->x - $v2->x, $v1->y - $v2->y, $v1->z - $v2->z );
  },
  '*' => sub {
    my ( $v1, $v2 ) = @_;
    if ( ref($v2) eq 'vec3' ) {
        return $v1->x * $v2->x + $v1->y * $v2->y + $v1->z * $v2->z;
    }
    else {
        return vec3->new( $v1->x * $v2, $v1->y * $v2, $v1->z * $v2 );
    }
  },
  '/' => sub {
    my ( $v1, $v2 ) = @_;
    return vec3->new( $v1->x / $v2, $v1->y / $v2, $v1->z / $v2 );
  },
  'x' => sub {
    my ( $v1, $v2 ) = @_;
    return vec3->new(
        $v1->y * $v2->z - $v1->z * $v2->y,
        $v1->z * $v2->x - $v1->x * $v2->z,
        $v1->x * $v2->y - $v1->y * $v2->x
    );
  },
  fallback => 1;

sub len {
    my $self = shift;
    return sqrt( $self * $self );
}

sub normalized {
    my $self = shift;
    my $d    = len($self);

    if ( $d == 0 ) {
        return vec3->new( 0, 0, 0 );
    }
    else {
        return $self / $d;
    }
}

sub as_string {
    my $self = shift;
    return join( ',', $self->x, $self->y, $self->z );
}
